@@ -50,7 +50,6 @@
 #define CompNumLockMask    (1 << 21)
 #define CompScrollLockMask (1 << 22)
 
-#define GNOME        "/desktop/gnome"
 #define METACITY     "/apps/metacity"
 #define COMPIZ       "/apps/compiz"
 #define COMPIZCONFIG "/apps/compizconfig"
@@ -78,11 +77,17 @@
 				 "%s/plugins/%s/%s/options/%s", COMPIZ, \
 				 setting->parent->name, keyName, setting->name);
 
+static const char* watchedGnomeDirectories[] = {
+    METACITY,
+    "/desktop/gnome/applications/terminal",
+    "/apps/panel/applets/window_list/prefs"
+};
+#define NUM_WATCHED_DIRS 3
+
 static GConfClient *client = NULL;
 static GConfEngine *conf = NULL;
-static guint backendNotifyId = 0;
-static guint metacityNotifyId = 0;
-static guint gnomeNotifyId = 0;
+static guint compizNotifyId;
+static guint gnomeNotifyIds[NUM_WATCHED_DIRS];
 static char *currentProfile = NULL;
 
 /* some forward declarations */
@@ -101,19 +106,23 @@ typedef enum {
     OptionSpecial,
 } SpecialOptionType;
 
-struct _SpecialOption {
+typedef struct _SpecialOption {
     const char*       settingName;
     const char*       pluginName;
     Bool	      screen;
     const char*       gnomeName;
     SpecialOptionType type;
-} const specialOptions[] = {
+} SpecialOption;
+
+const SpecialOption specialOptions[] = {
     {"run", "core", FALSE,
      METACITY "/global_keybindings/panel_run_dialog", OptionKey},
     {"main_menu", "core", FALSE,
      METACITY "/global_keybindings/panel_main_menu", OptionKey},
     {"run_command_screenshot", "core", FALSE,
      METACITY "/global_keybindings/run_command_screenshot", OptionKey},
+    {"run_command_window_screenshot", "core", FALSE,
+     METACITY "/global_keybindings/run_command_window_screenshot", OptionKey},
     {"run_command_terminal", "core", FALSE,
      METACITY "/global_keybindings/run_command_terminal", OptionKey},
 
@@ -125,9 +134,9 @@ struct _SpecialOption {
      METACITY "/window_keybindings/maximize", OptionKey},
     {"unmaximize_window", "core", FALSE,
      METACITY "/window_keybindings/unmaximize", OptionKey},
-    {"toggle_window_maximized_horizontally", "core", FALSE,
+    {"maximize_window_horizontally", "core", FALSE,
      METACITY "/window_keybindings/maximize_horizontally", OptionKey},
-    {"toggle_window_maximized_vertically", "core", FALSE,
+    {"maximize_window_vertically", "core", FALSE,
      METACITY "/window_keybindings/maximize_vertically", OptionKey},
     {"raise_window", "core", FALSE,
      METACITY "/window_keybindings/raise", OptionKey},
@@ -151,11 +160,21 @@ struct _SpecialOption {
     {"mouse_button_modifier", NULL, FALSE,
      METACITY "/general/mouse_button_modifier", OptionSpecial},
 
+    {"visual_bell", "fade", TRUE,
+     METACITY "/general/visual_bell", OptionBool},
+    {"fullscreen_visual_bell", "fade", TRUE,
+     METACITY "/general/visual_bell_type", OptionSpecial},
+
     {"next", "switcher", FALSE,
      METACITY "/global_keybindings/switch_windows", OptionKey},
     {"prev", "switcher", FALSE,
      METACITY "/global_keybindings/switch_windows_backward", OptionKey},
 
+    {"toggle_sticky", "extrawm", FALSE,
+     METACITY "/window_keybindings/toggle_on_all_workspaces", OptionKey},
+    {"toggle_fullscreen", "extrawm", FALSE,
+     METACITY "/window_keybindings/toggle_fullscreen", OptionKey},
+
     {"command0", "core", FALSE,
      METACITY "/keybinding_commands/command_1", OptionString},
     {"command1", "core", FALSE,
@@ -261,6 +280,31 @@ struct _SpecialOption {
     {"plane_to_12", "plane", FALSE,
      METACITY "/global_keybindings/switch_to_workspace_12", OptionKey},
 
+    {"switch_to_1", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_1", OptionKey},
+    {"switch_to_2", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_2", OptionKey},
+    {"switch_to_3", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_3", OptionKey},
+    {"switch_to_4", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_4", OptionKey},
+    {"switch_to_5", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_5", OptionKey},
+    {"switch_to_6", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_6", OptionKey},
+    {"switch_to_7", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_7", OptionKey},
+    {"switch_to_8", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_8", OptionKey},
+    {"switch_to_9", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_9", OptionKey},
+    {"switch_to_10", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_10", OptionKey},
+    {"switch_to_11", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_11", OptionKey},
+    {"switch_to_12", "vpswitch", FALSE,
+     METACITY "/global_keybindings/switch_to_workspace_12", OptionKey},
+
     {"up", "wall", FALSE,
      METACITY "/global_keybindings/switch_to_workspace_up", OptionKey},
     {"down", "wall", FALSE,
@@ -269,6 +313,31 @@ struct _SpecialOption {
      METACITY "/global_keybindings/switch_to_workspace_left", OptionKey},
     {"right", "wall", FALSE,
      METACITY "/global_keybindings/switch_to_workspace_right", OptionKey},
+    {"left_window", "wall", FALSE,
+     METACITY "/window_keybindings/move_to_workspace_left", OptionKey},
+    {"right_window", "wall", FALSE,
+     METACITY "/window_keybindings/move_to_workspace_right", OptionKey},
+    {"up_window", "wall", FALSE,
+     METACITY "/window_keybindings/move_to_workspace_up", OptionKey},
+    {"down_window", "wall", FALSE,
+     METACITY "/window_keybindings/move_to_workspace_down", OptionKey},
+
+    {"put_topleft", "put", FALSE,
+     METACITY "/window_keybindings/move_to_corner_nw", OptionKey},
+    {"put_topright", "put", FALSE,
+     METACITY "/window_keybindings/move_to_corner_ne", OptionKey},
+    {"put_bottomleft", "put", FALSE,
+     METACITY "/window_keybindings/move_to_corner_sw", OptionKey},
+    {"put_bottomright", "put", FALSE,
+     METACITY "/window_keybindings/move_to_corner_se", OptionKey},
+    {"put_left", "put", FALSE,
+     METACITY "/window_keybindings/move_to_side_w", OptionKey},
+    {"put_right", "put", FALSE,
+     METACITY "/window_keybindings/move_to_side_e", OptionKey},
+    {"put_top", "put", FALSE,
+     METACITY "/window_keybindings/move_to_side_n", OptionKey},
+    {"put_bottom", "put", FALSE,
+     METACITY "/window_keybindings/move_to_side_s", OptionKey},
 
     {"rotate_to_1_window", "rotate", FALSE,
      METACITY "/window_keybindings/move_to_workspace_1", OptionKey},
@@ -305,7 +374,11 @@ struct _SpecialOption {
     {"command_window_screenshot", "core", FALSE,
      METACITY "/keybinding_commands/command_window_screenshot", OptionString},
     {"command_terminal", "core", FALSE,
-     GNOME "/applications/terminal/exec", OptionString},
+     "/desktop/gnome/applications/terminal/exec", OptionString},
+
+    {"current_viewport", "thumbnail", TRUE,
+     "/apps/panel/applets/window_list/prefs/display_all_workspaces",
+     OptionSpecial},
 
     {"autoraise", "core", FALSE,
      METACITY "/general/auto_raise", OptionBool},
@@ -318,8 +391,8 @@ struct _SpecialOption {
 
     {"audible_bell", "core", FALSE,
      METACITY "/general/audible_bell", OptionBool},
-    {"hsize", "core", TRUE,
-     METACITY "/general/num_workspaces", OptionInt},
+    /*{"hsize", "core", TRUE,
+     METACITY "/general/num_workspaces", OptionInt},*/
 };
 
 #define N_SOPTIONS (sizeof (specialOptions) / sizeof (struct _SpecialOption))
@@ -441,9 +514,11 @@ valueChanged (GConfClient *client,
 
 	    if (strcmp (token + len - prefixLen, prefix[i]) == 0)
 	    {
-		char *buffer = strndup (token, len - prefixLen);
+		char *buffer = malloc (sizeof (char) * (len - prefixLen + 1));
 		if (buffer)
 		{
+		    strncpy (buffer, token, len - prefixLen);
+		    buffer[len - prefixLen] = 0;
 		    setting = ccsFindSetting (plugin, buffer,
 					      isScreen, screenNum);
 		    free (buffer);
@@ -475,80 +550,90 @@ gnomeValueChanged (GConfClient *client,
 {
     CCSContext *context = (CCSContext *)user_data;
     char       *keyName = (char*) gconf_entry_get_key (entry);
-    int        i, num = -1;
+    int        i, last = 0, num = 0;
+    Bool       needInit = TRUE;
 
     if (!ccsGetIntegrationEnabled (context))
 	return;
 
-    for (i = 0; i < N_SOPTIONS; i++)
+    /* we have to loop multiple times here, because one Gnome
+       option may be integrated with multiple Compiz options */
+
+    while (1)
     {
-	if (strcmp (specialOptions[i].gnomeName, keyName) == 0)
+	for (i = last, num = -1; i < N_SOPTIONS; i++)
 	{
-	    num = i;
-	    break;
-	}
-    }
-
-    if (num < 0)
-	return;
-
-    if (strcmp (specialOptions[num].settingName, "mouse_button_modifier") == 0)
-    {
-	CCSSetting *s;
-
-	readInit (context);
-
-	s = findDisplaySettingForPlugin (context, "core", "window_menu");
-	if (s)
-	    readSetting (context, s);
-
-	s = findDisplaySettingForPlugin (context, "move", "initiate");
-	if (s)
-	    readSetting (context, s);
-
-	s = findDisplaySettingForPlugin (context, "resize", "initiate");
-	if (s)
-	    readSetting (context, s);
-    }
-    else
-    {
-	CCSPlugin  *plugin = NULL;
-	CCSSetting *setting;
-	Bool       needInit = TRUE;
-
-	plugin = ccsFindPlugin (context,
-				(char*) specialOptions[num].pluginName);
-
-	if (!plugin)
-    	    return;
-
-	for (i = 0; i < context->numScreens; i++)
-	{
-	    unsigned int screen;
-
-	    if (specialOptions[num].screen)
-		screen = context->screens[i];
-	    else
-		screen = 0;
-
-	    setting = ccsFindSetting (plugin,
-				      (char*) specialOptions[num].settingName,
-				      specialOptions[num].screen, screen);
-
-	    if (setting)
+	    if (strcmp (specialOptions[i].gnomeName, keyName) == 0)
 	    {
-		if (needInit)
-		{
-		    readInit (context);
-    		    needInit = FALSE;
-		}
-		readSetting (context, setting);
+		num = i;
+		last = i + 1;
+		break;
+	    }
+	}
+
+	if (num < 0)
+	    break;
+
+	if (strcmp (specialOptions[num].settingName,
+		    "mouse_button_modifier") == 0)
+	{
+	    CCSSetting *s;
+
+	    if (needInit)
+	    {
+		readInit (context);
+		needInit = FALSE;
 	    }
 
-	    /* do not read display settings multiple
-	       times for multiscreen environments */
-	    if (!specialOptions[num].screen)
-		break;
+	    s = findDisplaySettingForPlugin (context, "core", "window_menu");
+	    if (s)
+		readSetting (context, s);
+
+	    s = findDisplaySettingForPlugin (context, "move", "initiate");
+	    if (s)
+		readSetting (context, s);
+
+	    s = findDisplaySettingForPlugin (context, "resize", "initiate");
+	    if (s)
+		readSetting (context, s);
+	}
+	else
+	{
+	    CCSPlugin     *plugin = NULL;
+	    CCSSetting    *setting;
+	    SpecialOption *opt = (SpecialOption *) &specialOptions[num];
+
+	    plugin = ccsFindPlugin (context, (char*) opt->pluginName);
+	    if (plugin)
+	    {
+		for (i = 0; i < context->numScreens; i++)
+		{
+		    unsigned int screen;
+
+		    if (opt->screen)
+			screen = context->screens[i];
+		    else
+			screen = 0;
+
+		    setting = ccsFindSetting (plugin, (char*) opt->settingName,
+					      opt->screen, screen);
+
+		    if (setting)
+		    {
+			if (needInit)
+			{
+			    readInit (context);
+			    needInit = FALSE;
+			}
+			readSetting (context, setting);
+		    }
+
+		    /* do not read display settings multiple
+		       times for multiscreen environments */
+		    if (!opt->screen)
+			i = context->numScreens;
+		}
+	    }
 	}
     }
 }
@@ -556,49 +641,47 @@ gnomeValueChanged (GConfClient *client,
 static void
 initClient (CCSContext *context)
 {
+    int i;
+
     client = gconf_client_get_for_engine (conf);
 
-    backendNotifyId = gconf_client_notify_add (client, COMPIZ,
-					       valueChanged, context,
-					       NULL, NULL);
-
-    metacityNotifyId = gconf_client_notify_add (client, METACITY,
-	   					gnomeValueChanged, context,
-	   					NULL,NULL);
-
-    gnomeNotifyId = gconf_client_notify_add (client,
-					     GNOME "/applications/terminal",
-      					     gnomeValueChanged, context,
-					     NULL,NULL);
-
+    compizNotifyId = gconf_client_notify_add (client, COMPIZ, valueChanged,
+					      context, NULL, NULL);
     gconf_client_add_dir (client, COMPIZ, GCONF_CLIENT_PRELOAD_NONE, NULL);
-    gconf_client_add_dir (client, METACITY, GCONF_CLIENT_PRELOAD_NONE, NULL);
-    gconf_client_add_dir (client, GNOME "/applications/terminal",
-			  GCONF_CLIENT_PRELOAD_NONE, NULL);
+
+    for (i = 0; i < NUM_WATCHED_DIRS; i++)
+    {
+	gnomeNotifyIds[i] = gconf_client_notify_add (client,
+						     watchedGnomeDirectories[i],
+						     gnomeValueChanged, context,
+						     NULL, NULL);
+	gconf_client_add_dir (client, watchedGnomeDirectories[i],
+			      GCONF_CLIENT_PRELOAD_NONE, NULL);
+    }
 }
 
 static void
 finiClient (void)
 {
-    if (backendNotifyId)
+    int i;
+
+    if (compizNotifyId)
     {
-	gconf_client_notify_remove (client, backendNotifyId);
-	backendNotifyId = 0;
+	gconf_client_notify_remove (client, compizNotifyId);
+	compizNotifyId = 0;
     }
-    if (metacityNotifyId)
+    gconf_client_remove_dir (client, COMPIZ, NULL);
+
+    for (i = 0; i < NUM_WATCHED_DIRS; i++)
     {
-	gconf_client_notify_remove (client, metacityNotifyId);
-	metacityNotifyId = 0;
-    }
-    if (gnomeNotifyId)
-    {
-	gconf_client_notify_remove (client, gnomeNotifyId);
-	gnomeNotifyId = 0;
+	if (gnomeNotifyIds[i])
+	{
+	    gconf_client_notify_remove (client, gnomeNotifyIds[0]);
+	    gnomeNotifyIds[i] = 0;
+	}
+	gconf_client_remove_dir (client, watchedGnomeDirectories[i], NULL);
     }
 
-    gconf_client_remove_dir (client, COMPIZ, NULL);
-    gconf_client_remove_dir (client, METACITY, NULL);
-    gconf_client_remove_dir (client, GNOME "/applications/terminal", NULL);
     gconf_client_suggest_sync (client, NULL);
 
     g_object_unref (client);
@@ -901,7 +984,10 @@ readListValue (CCSSetting *setting,
 
     valueList = gconf_value_get_list (gconfValue);
     if (!valueList)
-	return FALSE;
+    {
+	ccsSetList (setting, NULL);
+	return TRUE;
+    }
 
     nItems = g_slist_length (valueList);
 
@@ -1082,7 +1168,15 @@ readIntegratedOption (CCSContext *context,
 		CCSSettingActionValue action;
 		memset (&action, 0, sizeof(CCSSettingActionValue));
 		ccsGetAction (setting, &action);
-		if (ccsStringToKeyBinding (value, &action))
+		if (strcasecmp (value, "disabled") == 0)
+		{
+		    /* clear out keybinding */
+		    action.keysym     = NoSymbol;
+		    action.keyModMask = 0;
+		    ccsSetAction (setting, action);
+		    ret = TRUE;
+		}
+		else if (ccsStringToKeyBinding (value, &action))
 		{
 		    ccsSetAction (setting, action);
 		    ret = TRUE;
@@ -1096,7 +1190,37 @@ readIntegratedOption (CCSContext *context,
 	    const char *settingName = specialOptions[index].settingName;
 	    const char *pluginName  = specialOptions[index].pluginName;
 
-	    if (strcmp (settingName, "click_to_focus") == 0)
+	    if (strcmp (settingName, "current_viewport") == 0)
+	    {
+		const char *name;
+		gboolean   showAll;
+
+		name = specialOptions[index].gnomeName;
+		showAll = gconf_client_get_bool (client, name, &err);
+		if (!err)
+		{
+		    ccsSetBool (setting, !showAll);
+		    ret = TRUE;
+		}
+	    }
+	    else if (strcmp (settingName, "fullscreen_visual_bell") == 0)
+	    {
+		const char *name;
+		char       *value;
+
+		name = specialOptions[index].gnomeName;
+		value = gconf_client_get_string (client, name, &err);
+		if (!err && value)
+		{
+		    Bool fullscreen;
+
+		    fullscreen = strcmp (value, "fullscreen") == 0;
+		    ccsSetBool (setting, fullscreen);
+		    ret = TRUE;
+		    g_free (value);
+		}
+	    }
+	    else if (strcmp (settingName, "click_to_focus") == 0)
 	    {
 		char       *focusMode;
 		const char *name;
@@ -1127,14 +1251,22 @@ readIntegratedOption (CCSContext *context,
 		{
 		    CCSSettingActionValue action;
 		    memset (&action, 0, sizeof(CCSSettingActionValue));
-    		    ccsGetAction (setting, &action);
-	    	    if (ccsStringToKeyBinding (value, &action))
+		    ccsGetAction (setting, &action);
+
+		    if (strcasecmp (value, "disabled") == 0)
+		    {
+			/* clear out keybinding */
+			action.keysym     = NoSymbol;
+			action.keyModMask = 0;
+		    }
+		    if ((strcasecmp (value, "disabled") == 0) ||
+			ccsStringToKeyBinding (value, &action))
 		    {
 			action.buttonModMask = getGnomeMouseButtonModifier ();
 			if (strcmp (settingName, "window_menu") == 0)
-    			    action.button = 3;
+			    action.button = 3;
 			else if (strcmp (pluginName, "resize") == 0)
-	    		    action.button = 2;
+			    action.button = 2;
 			else
 			    action.button = 1;
 
@@ -1573,6 +1705,13 @@ writeIntegratedOption (CCSContext *context,
 	    gchar *currentValue;
 
 	    newValue = ccsKeyBindingToString (&setting->value->value.asAction);
+	    if (newValue && !strlen (newValue))
+	    {
+		free (newValue);
+		newValue = NULL;
+	    }
+	    if (!newValue)
+		newValue = strdup ("disabled");
 	    if (newValue)
     	    {
 		currentValue = gconf_client_get_string(client,
@@ -1594,7 +1733,35 @@ writeIntegratedOption (CCSContext *context,
 	    const char *settingName = specialOptions[index].settingName;
 	    const char *pluginName  = specialOptions[index].pluginName;
 
-	    if (strcmp (settingName, "click_to_focus") == 0)
+	    if (strcmp (settingName, "current_viewport") == 0)
+	    {
+		Bool currentViewport;
+
+		if (!ccsGetBool (setting, &currentViewport))
+		    break;
+
+		gconf_client_set_bool (client, optionName,
+				       !currentViewport, NULL);
+	    }
+	    else if (strcmp (settingName, "fullscreen_visual_bell") == 0)
+	    {
+		Bool  fullscreen;
+		gchar *currentValue, *newValue;
+		if (!ccsGetBool (setting, &fullscreen))
+		    break;
+
+		newValue = fullscreen ? "fullscreen" : "frame_flash";
+		currentValue = gconf_client_get_string (client,
+							optionName, &err);
+		if (!err && currentValue)
+		{
+		    if (strcmp (currentValue, newValue) != 0)
+			gconf_client_set_string (client, optionName,
+						 newValue, NULL);
+		    g_free (currentValue);
+		}
+	    }
+	    else if (strcmp (settingName, "click_to_focus") == 0)
 	    {
 		Bool  clickToFocus;
 		gchar *newValue, *currentValue;
@@ -1602,14 +1769,14 @@ writeIntegratedOption (CCSContext *context,
 		    break;
 
 		newValue = clickToFocus ? "click" : "mouse";
-		currentValue = gconf_client_get_string(client,
-						       optionName, &err);
+		currentValue = gconf_client_get_string (client,
+							optionName, &err);
 
 		if (!err && currentValue)
 		{
-		    if (strcmp(currentValue, newValue) != 0)
+		    if (strcmp (currentValue, newValue) != 0)
 			gconf_client_set_string (client, optionName,
-						 newValue,NULL);
+						 newValue, NULL);
 		    g_free (currentValue);
 		}
 	    }
@@ -1625,6 +1792,13 @@ writeIntegratedOption (CCSContext *context,
 		modMask = setting->value->value.asAction.buttonModMask;
 		newValue =
 		    ccsKeyBindingToString (&setting->value->value.asAction);
+		if (newValue && !strlen (newValue))
+		{
+		    free (newValue);
+		    newValue = NULL;
+		}
+		if (!newValue)
+		    newValue = strdup ("disabled");
 		if (newValue)
 		{
 		    currentValue = gconf_client_get_string(client,
@@ -1956,8 +2130,7 @@ writeSetting (CCSContext *context,
     {
 	writeIntegratedOption (context, setting, index);
     }
-    else if (setting->isDefault &&
-	     (strcmp (setting->name, "____plugin_enabled") != 0))
+    else if (setting->isDefault)
     {
 	resetOptionToDefault (setting);
     }
